#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_protocol.h"
#include "apr_buckets.h"
#include "apr_date.h"
#include "string.h"
#include "glob.h"
#include "sys/stat.h"

#define AIRBROAD_HANDLER "file-stream"
#define MAX_BUCKET_SIZE (0x3200000)

static apr_off_t fun_chunk_size(int res_int)
{
	if (res_int == -1 || res_int >= 1080)
	{
		return 2621440;
	}
	else if (res_int >= 720)
	{
		return 1572864;
	}
	else if (res_int >= 480)
	{
		return 1048576;
	}
	else
	{
		return 614400;
	}
}

static int file_stream(request_rec *r)
{
	apr_bucket_brigade *bb;
	apr_file_t *fp = NULL;
	apr_off_t chunk_size;
	size_t i;
	int res_int;
	int rc;

	if (!r->args)
	{
		return DECLINED;
	}

	if (r->args[0] == '\0')
	{
		return HTTP_BAD_REQUEST;
	}

	if ((rc = ap_discard_request_body(r)) != OK)
	{
		return rc;
	}

	if ((rc = apr_file_open(&fp, r->filename, APR_READ, APR_OS_DEFAULT, r->pool)))
	{
		return rc;
	}

	bb = apr_brigade_create(r->pool, r->connection->bucket_alloc);

	if (r->args[0] == 'r' && r->args[1] == '\0')
	{
		const char *modified;

		if ((modified = apr_table_get(r->headers_in, "If-Modified-Since")))
		{
			apr_time_t since_mtime = apr_date_parse_rfc(modified);

			if (since_mtime + 18000000000 <= r->request_time)
			{
				if (since_mtime + 21600000000 >= r->request_time)
				{
					return HTTP_NOT_MODIFIED;
				}
			}
			else if (since_mtime + 16000000 > r->request_time)
			{
				return HTTP_NOT_MODIFIED;
			}
		}

		char file_info[12];
		glob_t globbuf;
		struct stat statbuf;
		size_t name_i = 0;
		char* path;
		char* extension;
		char* resolution;
		int len = 0;
		apr_time_t mtime = 0;
		
		extension = (char *)malloc(sizeof(char *) * 4);
		extension[3] = '\0';
		i = strlen(r->finfo.fname) - 1;
		extension[2] = r->finfo.fname[i--];
		extension[1] = r->finfo.fname[i--];
		extension[0] = r->finfo.fname[i];
		i = i - 2;
		
		resolution = (char *)malloc(sizeof(char *));

		path = (char *)malloc(sizeof(char *) * (strlen(r->finfo.fname) + 3));
		strcpy(path, r->finfo.fname);

		if (path[i] == 'p')
		{
			i--;

			if (path[i] >= '0' && path[i] <= '9')
			{
				resolution[0] = path[i];
				i--;
				len++;

				while (path[i] >= '0' && path[i] <= '9')
				{
					resolution = (char *)realloc(resolution, sizeof(char *) * (len + 1));
					resolution[len] = path[i];
					i--;
					len++;
				}

				resolution = (char *)realloc(resolution, sizeof(char *) * (len + 1));
				resolution[len] = '\0';

				if (path[i] == '_' && path[i+1] != '0')
				{
					name_i = i;
				}
			}
			else
			{
				resolution[0] = '\0';
			}
		}
		else
		{
			resolution[0] = '\0';
		}

		if (name_i == 0)
		{
			name_i = strlen(r->finfo.fname) - 4;
			i = name_i;
		}

		path[i++] = '.';
		path[i++] = 'f';
		path[i++] = 'l';
		path[i++] = 'v';
		path[i] = '\0';

		memset(&globbuf, 0, sizeof(glob_t));

		if (glob(path, GLOB_NOSORT, NULL, &globbuf) == 0)
		{
			stat(globbuf.gl_pathv[0], &statbuf);
			sprintf(file_info, "%lld", (long long) statbuf.st_size);
			apr_brigade_write(bb, NULL, NULL, file_info, strlen(file_info));

			chunk_size = fun_chunk_size(-1);
			sprintf(file_info, ",%lld,", (long long) chunk_size);
			apr_brigade_write(bb, NULL, NULL, file_info, strlen(file_info));

			if (extension[0] == 'f' && extension[1] == 'l' && extension[2] == 'v' && len == 0)
			{
				apr_brigade_write(bb, NULL, NULL, "0a,", 3);
				len = -1;
			}
			else
			{
				apr_brigade_write(bb, NULL, NULL, "0,", 2);
			}

			i = strlen(r->uri) - 1;

			while (r->uri[i] != '/')
			{
				i--;
			}

			i++;
			apr_brigade_write(bb, NULL, NULL, r->uri, i);

			i = strlen(globbuf.gl_pathv[0]) - 5;

			while (globbuf.gl_pathv[0][i] != '/')
			{
				i--;
			}

			i++;
			
			while (globbuf.gl_pathv[0][i] != '\0')
			{
				apr_brigade_write(bb, NULL, NULL, &(globbuf.gl_pathv[0][i]), 1);
				
				i++;
			}

			apr_brigade_write(bb, NULL, NULL, ",", 1);

			mtime = (long long)statbuf.st_mtime;
		}

		globfree(&globbuf);

		i = name_i;

		path[i++] = '_';
		path[i++] = '*';
		path[i++] = 'p';
		path[i++] = '.';
		path[i++] = 'f';
		path[i++] = 'l';
		path[i++] = 'v';
		path[i] = '\0';
		
		memset(&globbuf, 0, sizeof(glob_t));

		if (glob(path, GLOB_NOSORT, NULL, &globbuf) == 0)
		{
			int num = 0;
			int j;

			do{
				i = strlen(globbuf.gl_pathv[num]) - 6;
				j = 0;

				if (globbuf.gl_pathv[num][i] >= '0' && globbuf.gl_pathv[num][i] <= '9')
				{
					i--;

					while (globbuf.gl_pathv[num][i] >= '0' && globbuf.gl_pathv[num][i] <= '9')
					{
						i--;
					}

					if (globbuf.gl_pathv[num][i] == '_' && globbuf.gl_pathv[num][i + 1] != '0')
					{
						size_t i_tmp = ++i;
						res_int = 0;

						while (globbuf.gl_pathv[num][i] >= '0' && globbuf.gl_pathv[num][i] <= '9')
						{
							res_int = res_int * 10 + (int)globbuf.gl_pathv[num][i++] - 48;

						}

						stat(globbuf.gl_pathv[num], &statbuf);
						sprintf(file_info, "%lld", (long long)statbuf.st_size);
						apr_brigade_write(bb, NULL, NULL, file_info, strlen(file_info));

						chunk_size = fun_chunk_size(res_int);
						sprintf(file_info, ",%lld,", (long long) chunk_size);
						apr_brigade_write(bb, NULL, NULL, file_info, strlen(file_info));

						while (i_tmp < i)
						{
							file_info[j++] = globbuf.gl_pathv[num][i_tmp++];
						}

						if (j == len && len > 0)
						{
							if (extension[0] == 'f' && extension[1] == 'l' && extension[2] == 'v')
							{
								int len_tmp = 0;

								while (len_tmp < len)
								{
									if (resolution[len_tmp++] != globbuf.gl_pathv[num][--i])
									{
										break;
									}
								}

								if (len_tmp == len)
								{
									file_info[j++] = 'a';
									len = -1;
								}
							}
						}

						file_info[j++] = ',';
						file_info[j] = '\0';
					}
				}

				if (j == 0)
				{
					continue;
				}

				apr_brigade_write(bb, NULL, NULL, file_info, strlen(file_info));

				i = strlen(r->uri) - 1;

				while (r->uri[i] != '/')
				{
					i--;
				}

				i++;
				apr_brigade_write(bb, NULL, NULL, r->uri, i);

				i = strlen(globbuf.gl_pathv[num]) - 8;

				while (globbuf.gl_pathv[num][i] != '/')
				{
					i--;
				}

				i++;

				while (globbuf.gl_pathv[num][i] != '\0')
				{
					apr_brigade_write(bb, NULL, NULL, &(globbuf.gl_pathv[num][i]), 1);

					i++;
				}

				apr_brigade_write(bb, NULL, NULL, ",", 1);

				if (mtime < statbuf.st_mtime)
				{
					mtime = (long long)statbuf.st_mtime;
				}

				num++;
			}
			while (num < globbuf.gl_pathc);
		}

		globfree(&globbuf);

		i = name_i;
		
		path[i++] = '.';
		path[i++] = 'm';
		path[i++] = 'p';
		path[i++] = '4';
		path[i] = '\0';

		memset(&globbuf, 0, sizeof(glob_t));

		if (glob(path, GLOB_NOSORT, NULL, &globbuf) == 0)
		{
			stat(globbuf.gl_pathv[0], &statbuf);
			sprintf(file_info, "%lld", (long long)statbuf.st_size);
			apr_brigade_write(bb, NULL, NULL, file_info, strlen(file_info));

			chunk_size = fun_chunk_size(-1);
			sprintf(file_info, ",%lld,", (long long) chunk_size);
			apr_brigade_write(bb, NULL, NULL, file_info, strlen(file_info));

			if (extension[0] == 'm' && extension[1] == 'p' && extension[2] == '4' && len == 0)
			{
				apr_brigade_write(bb, NULL, NULL, "0a,", 3);
			}
			else
			{
				apr_brigade_write(bb, NULL, NULL, "0,", 2);
			}

			i = strlen(r->uri) - 1;

			while (r->uri[i] != '/')
			{
				i--;
			}

			i++;
			apr_brigade_write(bb, NULL, NULL, r->uri, i);

			i = strlen(globbuf.gl_pathv[0]) - 5;

			while (globbuf.gl_pathv[0][i] != '/')
			{
				i--;
			}

			i++;

			while (globbuf.gl_pathv[0][i] != '\0')
			{
				apr_brigade_write(bb, NULL, NULL, &(globbuf.gl_pathv[0][i]), 1);

				i++;
			}

			apr_brigade_write(bb, NULL, NULL, ",", 1);

			if (mtime < statbuf.st_mtime)
			{
				mtime = (long long)statbuf.st_mtime;
			}
		}

		globfree(&globbuf);

		i = name_i;

		path[i++] = '_';
		path[i++] = '*';
		path[i++] = 'p';
		path[i++] = '.';
		path[i++] = 'm';
		path[i++] = 'p';
		path[i++] = '4';
		path[i] = '\0';

		memset(&globbuf, 0, sizeof(glob_t));

		if (glob(path, GLOB_NOSORT, NULL, &globbuf) == 0)
		{
			int num = 0;
			int j;

			do{
				i = strlen(globbuf.gl_pathv[num]) - 6;
				j = 0;

				if (globbuf.gl_pathv[num][i] >= '0' && globbuf.gl_pathv[num][i] <= '9')
				{
					i--;
					
					while (globbuf.gl_pathv[num][i] >= '0' && globbuf.gl_pathv[num][i] <= '9')
					{
						i--;
					}

					if (globbuf.gl_pathv[num][i] == '_' && globbuf.gl_pathv[num][i + 1] != '0')
					{
						size_t i_tmp = ++i;
						res_int = 0;

						while (globbuf.gl_pathv[num][i] >= '0' && globbuf.gl_pathv[num][i] <= '9')
						{
							res_int = res_int * 10 + (int)globbuf.gl_pathv[num][i++] - 48;

						}

						stat(globbuf.gl_pathv[num], &statbuf);
						sprintf(file_info, "%lld", (long long)statbuf.st_size);
						apr_brigade_write(bb, NULL, NULL, file_info, strlen(file_info));

						chunk_size = fun_chunk_size(res_int);
						sprintf(file_info, ",%lld,", (long long) chunk_size);
						apr_brigade_write(bb, NULL, NULL, file_info, strlen(file_info));

						while (i_tmp < i)
						{
							file_info[j++] = globbuf.gl_pathv[num][i_tmp++];
						}

						if (j == len && len > 0)
						{
							if (extension[0] == 'm' && extension[1] == 'p' && extension[2] == '4')
							{
								int len_tmp = 0;

								while (len_tmp < len)
								{
									if (resolution[len_tmp++] != globbuf.gl_pathv[num][--i])
									{
										break;
									}	
								}

								if (len_tmp == len)
								{
									file_info[j++] = 'a';
									len = -1;
								}
							}
						}

						file_info[j++] = ',';
						file_info[j] = '\0';
					}
				}

				if (j == 0)
				{
					continue;
				}

				apr_brigade_write(bb, NULL, NULL, file_info, strlen(file_info));

				i = strlen(r->uri) - 1;

				while (r->uri[i] != '/')
				{
					i--;
				}

				i++;
				apr_brigade_write(bb, NULL, NULL, r->uri, i);

				i = strlen(globbuf.gl_pathv[num]) - 8;

				while (globbuf.gl_pathv[num][i] != '/')
				{
					i--;
				}

				i++;

				while (globbuf.gl_pathv[num][i] != '\0')
				{
					apr_brigade_write(bb, NULL, NULL, &(globbuf.gl_pathv[num][i]), 1);

					i++;
				}

				apr_brigade_write(bb, NULL, NULL, ",", 1);

				if (mtime < statbuf.st_mtime)
				{
					mtime = (long long)statbuf.st_mtime;
				}

				num++;
			}
			while (num < globbuf.gl_pathc);
		}

		globfree(&globbuf);

		free(path);
		free(resolution);
		free(extension);
		
		mtime = mtime * 1000000;

		if (mtime + 18000000000 <= r->request_time)
		{
			r->mtime = r->request_time - 18000000000;
			apr_table_set(r->headers_out, "Cache-Control", "max-age=86400");
		}
		else
		{
			r->mtime = r->request_time;
			apr_table_set(r->headers_out, "Cache-Control", "max-age=300");
		}

		ap_set_last_modified(r);

		return ap_pass_brigade(r->output_filters, bb);
	}
	else
	{
		const char *modified;

		if ((modified = apr_table_get(r->headers_in, "If-Modified-Since")))
		{
			if (apr_date_parse_rfc(modified) + 1000000 > r->finfo.mtime)
			{
				return HTTP_NOT_MODIFIED;
			}
		}

		res_int = -1;
		i = strlen(r->finfo.fname) - 5;

		if (r->finfo.fname[i] == 'p')
		{
			i--;

			if (r->finfo.fname[i] >= '0' && r->finfo.fname[i] <= '9')
			{
				i--;

				while (r->finfo.fname[i] >= '0' && r->finfo.fname[i] <= '9')
				{
					i--;
				}

				if (r->finfo.fname[i] == '_' && r->finfo.fname[i + 1] != '0')
				{
					i++;
					res_int = 0;

					while (r->finfo.fname[i] >= '0' && r->finfo.fname[i] <= '9')
					{
						res_int = res_int * 10 + (int)r->finfo.fname[i] - 48;
						i++;
					}
				}
			}
		}

		chunk_size = fun_chunk_size(res_int);

		apr_off_t start;
		apr_off_t number;
		apr_off_t length;

		start = 0;
		number = 0;
		length = 0;
		i = 0;

		for (; r->args[i] != '\0'; i++)
		{
			switch (r->args[i]) {
			case '1':
				number = number * 10 + 1;
				break;
			case '2':
				number = number * 10 + 2;
				break;
			case '3':
				number = number * 10 + 3;
				break;
			case '4':
				number = number * 10 + 4;
				break;
			case '5':
				number = number * 10 + 5;
				break;
			case '6':
				number = number * 10 + 6;
				break;
			case '7':
				number = number * 10 + 7;
				break;
			case '8':
				number = number * 10 + 8;
				break;
			case '9':
				number = number * 10 + 9;
				break;
			case '0':
				number = number * 10;
				break;
			case 'r':
				if (length == 0)
				{
					start = number;
					number = 0;
					length = 1;
				}
				else return HTTP_BAD_REQUEST;
				break;
			default:
				return HTTP_BAD_REQUEST;
			}

			if (number >= r->finfo.size) return HTTP_BAD_REQUEST;
		}

		if (length == 0)
		{
			start = chunk_size * number;

			if (start + chunk_size > r->finfo.size)
			{
				length = r->finfo.size - start;
			}
			else
			{
				length = chunk_size;
			}
		}
		else
		{
			if (r->args[i - 1] == 'r')
			{
				length = r->finfo.size - start;
			}
			else
			{
				if (number < r->finfo.size)
				{
					if (start <= number)
					{
						length = number - start + 1;
					}
					else
					{
						return HTTP_BAD_REQUEST;
					}
				}
				else
				{
					return HTTP_BAD_REQUEST;
				}
			}
		}

		if (!apr_table_get(r->headers_in, "x-flash-version"))
		{
			const char *range;

			if ((range = apr_table_get(r->headers_in, "Range")))
			{
				int send_range = 1;
				const char *if_range;

				if ((if_range = apr_table_get(r->headers_in, "If-Range")))
				{
					if (apr_date_parse_rfc(if_range) + 1000000 < r->finfo.mtime)
					{
						send_range = 0;
					}
				}
				else
				{
					const char *unmodified;

					if ((unmodified = apr_table_get(r->headers_in, "If-Unmodified-Since")))
					{
						if (apr_date_parse_rfc(unmodified) + 1000000 < r->finfo.mtime)
						{
							return HTTP_PRECONDITION_FAILED;
						}
					}
				}

				if (send_range)
				{
					if (strncasecmp(range, "bytes=", 6) == 0)
					{
						size_t range_len = strlen(range);
						size_t offset = 6;
						apr_off_t begin = 0;
						apr_off_t end = 0;
						char str[50];

						while (offset < range_len)
						{
							if (range[offset] >= '0' && range[offset] <= '9')
							{
								begin = begin * 10 + (apr_off_t)(range[offset] - '0');
							}
							else if (range[offset] == '-')
							{
								offset++;
								break;
							}
							else
							{
								sprintf(str, "bytes */%lld", (long long)length);
								apr_table_set(r->headers_out, "Content-Range", str);
								return HTTP_RANGE_NOT_SATISFIABLE;
							}

							offset++;
						}

						if (offset < range_len)
						{
							do
							{
								if (range[offset] >= '0' && range[offset] <= '9')
								{
									end = end * 10 + (apr_off_t)(range[offset] - '0');
								}
								else if (range[offset] == ',')
								{
									break;
								}
								else
								{
									sprintf(str, "bytes */%lld", (long long)length);
									apr_table_set(r->headers_out, "Content-Range", str);
									return HTTP_RANGE_NOT_SATISFIABLE;
								}

								offset++;
							} while (offset < range_len);
						}
						else
						{
							end = length - 1;
						}

						if (begin > end || end >= length)
						{
							sprintf(str, "bytes */%lld", (long long)length);
							apr_table_set(r->headers_out, "Content-Range", str);
							return HTTP_RANGE_NOT_SATISFIABLE;
						}

						sprintf(str, "bytes %lld-%lld/%lld", (long long)begin, (long long)end, (long long)length);
						length = end - begin + 1;
						start += begin;
						apr_table_set(r->headers_out, "Content-Range", str);
						r->status = HTTP_PARTIAL_CONTENT;
						r->status_line = "206 Partial Content";
					}
				}
			}
		}

		r->mtime = r->finfo.mtime;
		ap_set_last_modified(r);
		ap_set_content_length(r, length);
		apr_table_set(r->headers_out, "Cache-Control", "max-age=432000");
		
		apr_bucket *e;

		if (sizeof(apr_off_t) == sizeof(apr_size_t) || length < MAX_BUCKET_SIZE)
		{
			e = apr_bucket_file_create(fp, start, (apr_size_t)length, r->pool, bb->bucket_alloc);
		}
		else
		{
			e = apr_bucket_file_create(fp, start, MAX_BUCKET_SIZE, r->pool, bb->bucket_alloc);

			while (length > MAX_BUCKET_SIZE)
			{
				apr_bucket *ce;

				apr_bucket_copy(e, &ce);
				APR_BRIGADE_INSERT_TAIL(bb, ce);
				e->start += MAX_BUCKET_SIZE;
				length -= MAX_BUCKET_SIZE;
			}

			e->length = (apr_size_t)length;
		}
		
		APR_BRIGADE_INSERT_TAIL(bb, e);
		
		return ap_pass_brigade(r->output_filters, bb);
	}
}

static int airbroad_handler(request_rec *r)
{
    if ((!r->handler) || (strcmp(r->handler, AIRBROAD_HANDLER)))
	{
        return DECLINED;
    }

    r->allowed |= (AP_METHOD_BIT << M_GET);

    if (r->method_number != M_GET)
	{
        return HTTP_METHOD_NOT_ALLOWED;
    }

    return file_stream(r);
}

static const command_rec airbroad_cmds[] =
{
    {NULL}
};

static void register_hooks(apr_pool_t *p)
{
    ap_hook_handler(airbroad_handler, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA airbroad_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,
    NULL,
    NULL,
    NULL,
    airbroad_cmds,
    register_hooks
};